PROCEDURE HOTKEY

* (c) 1990,1991 by Robert J. Grommes dba B & G Microsystems
*                                        3645 - 28th SE, Suite 236
*                                        Grand Rapids, MI  49512-1696
*                                        (616) 956-6022
*                                        CompuServe: 74126,72
* Released to the public domain 05/12/91 for private and commercial use,
* provided that this comment block with copyright and authorship remains in
* the source code.

* Calls: HOTKEYCL

* Hot key manager.  Use to register, save and restore hot key assignments.
* 1st parm is the function keyword - one of REGISTER, RELEASE, SAVE or RESTORE.
* These keywords are allowed to be abbreviated down to 4 characters.

*    REGISTER:  activates the hot key indicated by parm 2 to perform action
*               specified in parm 3, and adds this key definition to the
*               current hot key register.
*    UNREGISTER:  deactivates the hot key indicated by parm 2 and removes it
*                 from the current hot key register.  No error is returned
*                 if the assignment to be RELEASEd is not, in fact, in effect.
*                 If parm 2 is not passed, all current hot key assignments are
*                 cleared without popping the assignment stack.
*    SAVE: deactivates all current hot key assignments, if any, and pushes
*          them onto a hot key stack.  REGISTER can then be called to set
*          up new key assignments.
*    RESTORE: deactivates all current hot key assignments, if any, pops
*             the last SAVEd assigments off the stack, if any, and
*             activates the popped assignments.

* PUBLIC variables used: _cHOT_KEYS(), a character array of sufficient
* depth to accomodate the hot key assigments stack.  Each element of the
* array is initially set to null.  In operation, HOTKEY stores pairs of key
* labels and assignments delimited with tildes.  For instance:

* "F10~KEYBOARD CHR(23) PLAIN~F12~DO A_POP_UP WITH nVAR1,nVAR2~"

* _nHOT_KEYS - the size of the hot key assigments stack
* _nHOT_KEY  - the hot key assignments stack pointer

* If the caller hasn't initialized these 3 public variables, HOTKEY sets up
* the variables with a default stack size of 6 -- which should be plenty
* for most applications.

* Hot key management facilities provided by HOTKEY are needed by generic
* routines that need to use hot key assignments without disturbing the
* caller's environment.  By registering and clearing all key assigments
* through HOTKEY, it is possible for any FoxPro procedure to save the caller's
* key assignments, clear them, set up its own, then restore the caller's
* assignments upon return.  This capability can be nested as deeply as
* necessary by adjusting the hot key stack size.

* You can have the equivalent of a "global" or "public" hot key assignment
* by assigning such keys directly without using HOTKEY.  Such assignments
* are unaffected by SAVE and RESTORE operations in HOTKEY.  For instance, you
* might want to have your help key defined in this way.  This is why we
* don't use the PUSH/POP KEY facility introduced in FP 2.0.

* It would be simple to trap for defining an ON KEY for a LABEL that's
* already in use, though this version doesn't do so.  If you want this to
* be an error, uncomment the commented lines in the CASE for "REGISTER".

* Sample invocations of HOTKEY:

* Set up a hot key, equivalent to ON KEY LABEL ENTER KEYBOARD CHR(23) PLAIN
* DO HOTKEY WITH 'REGISTER','ENTER','KEYBOARD CHR(23) PLAIN'
* Clear all hot keys, saving them on stack
* DO HOTKEY WITH 'SAVE'
* Clear all hot keys, then restore hot key assignments previously saved on
* stack, if any.
* DO HOTKEY WITH 'RESTORE'
* Clear a hot key -- equivalent to ON KEY LABEL ENTER
* DO HOTKEY WITH 'UNREGISTER','ENTER'


PARAMETERS cFUNC,cKEY,cACTION
IF TYPE('_nHOT_KEYS') <> 'N'
 PUBLIC _nHOT_KEYS
 _nHOT_KEYS = 6
ENDIF
IF TYPE('_cHOT_KEYS(1)') <> 'C'
 PUBLIC _nHOT_KEY, _cHOT_KEYS(_nHOT_KEYS)
 _nHOT_KEY = 1
 _cHOT_KEYS = ''
ENDIF
cFUNC = ALLTRIM(UPPER(cFUNC))
DO CASE
 CASE cFUNC = LEFT('REGISTER',MAX(LEN(cFUNC),4))
  cKEY = ALLTRIM(UPPER(cKEY))
  cACTION = ALLTRIM(UPPER(cACTION))
  * Uncomment following lines if you want the definition of an already-defined
  * key label to be an error.  You could also use the ONKEY() function to
  * trap for "global" hot keys not defined via HOTKEY.
  * IF cKEY + '~' + cACTION + '~' $ _cHOT_KEYS(_nHOT_KEY)
  *  <your error handler here>
  * ENDIF
  _cHOT_KEYS(_nHOT_KEY) = _cHOT_KEYS(_nHOT_KEY) + cKEY + '~' + cACTION + '~'
  ON KEY LABEL &cKEY &cACTION
 CASE cFUNC = LEFT('UNREGISTER',MAX(LEN(cFUNC),4))
  PRIVATE n
  cKEY = ALLTRIM(UPPER(cKEY))
  * Get pointer to start of key
  IF LEFT(_cHOT_KEYS(_nHOT_KEY),LEN(cKEY) + 1) = cKEY + '~'
   n = 1
  ELSE
   n = AT('~' + cKEY + '~',_cHOT_KEYS(_nHOT_KEY)) + 1
  ENDIF
  IF n > 0
   * Get portion of current register starting with desired key label
   cACTION = SUBSTR(_cHOT_KEYS(_nHOT_KEY),n)
   * Throw away everything before start of desired action except delimiter
   cACTION = SUBSTR(cACTION,AT('~',cACTION))
   * Lop off everything following desired action except delimiter
   cACTION = LEFT(cACTION,AT('~',SUBSTR(cACTION,2)) + 1)
   * Deactivate the key assignment
   ON KEY LABEL &cKEY
   * Remove the key assignment (e.g. "KEY~ACTION~") from the register
   _cHOT_KEYS(_nHOT_KEY) = STUFF(_cHOT_KEYS(_nHOT_KEY),n,LEN(cKEY + cACTION),'')
  ENDIF
 CASE cFUNC = 'SAVE'
  IF _nHOT_KEY < _nHOT_KEYS
   DO HOTKEYCL
   _nHOT_KEY = _nHOT_KEY + 1
  ENDIF
 CASE cFUNC = LEFT('RESTORE',MAX(LEN(cFUNC),4))
  DO HOTKEYCL
  _cHOT_KEYS(_nHOT_KEY) = ''
  IF _nHOT_KEY > 1
   _nHOT_KEY = _nHOT_KEY - 1
   IF .NOT. EMPTY(_cHOT_KEYS(_nHOT_KEY))
    * Restore old hot key assignments
    PRIVATE c,n
    c = _cHOT_KEYS(_nHOT_KEY)
    n = AT('~',c)
    DO WHILE n > 1
     cLABEL = LEFT(c,n - 1)
     c = SUBSTR(c,n + 1)
     n = AT('~',c)
     cCMD = LEFT(c,n - 1)
     ON KEY LABEL &cLABEL &cCMD
     c = SUBSTR(c,n + 1)
     n = AT('~',c)
    ENDDO
   ENDIF
  ENDIF
ENDCASE
RETURN

PROCEDURE HOTKEYCL

* Called by HOTKEY.  Clears current hot key assignments.

PRIVATE c,n,cLABEL
c = _cHOT_KEYS(_nHOT_KEY)
n = AT('~',c)
DO WHILE n > 1
 cLABEL = LEFT(c,n - 1)
 ON KEY LABEL &cLABEL
 c = SUBSTR(c,n + 1)
 n = AT('~',c)
 c = SUBSTR(c,n + 1)
 n = AT('~',c)
ENDDO
RETURN

PROCEDURE FLDMOUSE
* This is an ON KEY assigned by GET handlers to LEFTMOUSE.  It does nothing
* if the mouse is being clicked in the current field, so that READ can
* handle mouse-directed intra-field cursor movements.  Otherwise it stuffs
* the keyboard with a keystroke to terminate the read.  Assuming the
* current field passes its edits, SETFLD will eventually be called and
* will move us to the field pointed to by the mouse.
IF .NOT. ;
       BETWEEN((MROW() * SCOLS()) + MCOL(),nMOUSE(nFIELD,1),nMOUSE(nFIELD,2))
 bFLDMOUSE = .t.
 KEYBOARD CHR(10)
ENDIF

FUNCTION ISSTATE

* Tests two upper-case characters passed against all valid U.S. state
* abbreviations, excluding territories, plus 'CN' for Canada.
* Returns true if valid, false otherwise.

PARAMETER cPARM
PRIVATE n
n = 1
DO WHILE n < 105
 IF SUBSTR('MIWIILINOHMNIAMOARLAKYTNMSALGAFLSCNCVAWVDCMDDEPANJNYCTRIMAVTNHMETXOKKSNESDNDNMCOWYMTIDUTAZNVCAORWAAKHICN',n,2) == cPARM
  * to cover US territories, add 'PRVITTGUCMTTAS' and increase loop test value.
  RETURN .t.
 ENDIF
 n = n + 2
ENDDO
RETURN .f.

PROCEDURE DISPREPT
* Report display handler.  Given the name of a report temp file and a
* window title, show the report in a non-editable full screen edit session,
* with option to print, file or discard the report on exit.  On EGA or VGA
* systems, the user can toggle between 25 and 43/50 line mode using [Alt-V].

* Call as follows:  DO DISPREPT WITH <expC1>,[<expC2>,],[<expC3>,],[<expC4>]
* Where:  <expC1> is the name of the file containing the report.  The file
*                 should contain no control codes except for form feeds.
*         <expC2> is an optional title for the report viewing window.
*                 If omitted or null, "VIEW REPORT" is used.
*         <expC3> is an optional control string to send to the printer at the
*                 start of printing, should the user elect to print the
*                 report.  If omitted or null, nothing is sent.
*         <expC4> is an optional control string to send to the printer at the
*                 end of printing, should the user elect to print the report.
*                 If omitted or null, nothing is sent.
* Note:  If both <expC3> AND <expC4> are omitted, the user is not offered
*        the print option.  If you want the user to be able to print but
*        don't want to send printer control codes, pass null strings for
*        <expC3> and <expC4>.
PARAMETERS cOUTFILE,cTITLE,cSTART,cEND
PRIVATE nNEWMODE,nOLDMODE,cFUNC,nHANDLE,cALERT_WIN,cFILE,cOLDPATH,cNEWPATH
IF PARAMETERS() = 3
 cEND = ''
ENDIF
IF PARAMETERS() < 3
 STORE '' TO cSTART,cEND
ENDIF
STORE 25 TO nNEWMODE,nOLDMODE
@ 3,0 CLEAR
IF PARAMETERS() < 3
 bPRINT = .f.
 @ 24,8 SAY 'Use mouse or arrow / paging keys to view; or press'
 @ 24,59 PROMPT '\<File or'
 @ 24,67 PROMPT '\<Quit'
ELSE
 bPRINT = .t.
 @ 24,5 SAY 'Use mouse or arrow / paging keys to view; or press'
 @ 24,56 PROMPT '\<Print,'
 @ 24,63 PROMPT '\<File, or'
 @ 24,72 PROMPT '\<Quit'
ENDIF
CLEAR PROMPT
DO WINCOLOR WITH 'SYSWINDOW'
IF EMPTY(cTITLE)
 cTITLE = 'TITLE [VIEW REPORT]'
ELSE
 cTITLE = 'TITLE [' + cTITLE + ']'
ENDIF
DEFINE WINDOW preview FROM 0,0 TO 23,79 SYSTEM CLOSE &cTITLE
cFUNC = ' '
IF bPRINT
 DO HOTKEY WITH 'REGISTER','P','cFUNC = DISPEXIT()'
ENDIF
DO HOTKEY WITH 'REGISTER','F','cFUNC = DISPEXIT()'
DO HOTKEY WITH 'REGISTER','Q','cFUNC = DISPEXIT()'
DO HOTKEY WITH 'REGISTER','ALT-V','nNEWMODE = VMODE_TOG()'
DO WHILE .t.
 MODIFY FILE (cOUTFILE) WINDOW preview NOEDIT
 IF nNEWMODE = nOLDMODE
  EXIT
 ENDIF
 RELEASE WINDOW preview
 DO WINCOLOR WITH 'NORMAL'
 ACTIVATE SCREEN
 @ 3,0 CLEAR
 IF bPRINT
  @ nNEWMODE - 1,5 SAY 'Use mouse or arrow / paging keys to view; or press'
  @ nNEWMODE - 1,56 PROMPT '\<Print,'
  @ nNEWMODE - 1,63 PROMPT '\<File, or'
  @ nNEWMODE - 1,72 PROMPT '\<Quit'
 ELSE
  @ nNEWMODE - 1,8 SAY 'Use mouse or arrow / paging keys to view; or press'
  @ nNEWMODE - 1,59 PROMPT '\<File, or'
  @ nNEWMODE - 1,67 PROMPT '\<Quit'
 ENDIF
 CLEAR PROMPT
 DO WINCOLOR WITH 'SYSWINDOW'
 DEFINE WINDOW preview FROM 0,0 TO nNEWMODE - 2,79 SYSTEM CLOSE &cTITLE
 nOLDMODE = nNEWMODE
ENDDO
RELEASE WINDOW preview
DO WINCOLOR WITH 'NORMAL'
ACTIVATE SCREEN
DO HOTKEY WITH 'UNREGISTER','ALT-V'
IF nOLDMODE > 25
 IF 'EGA' $ UPPER(SYS(2006))
  SET DISPLAY TO EGA25
 ELSE
  SET DISPLAY TO VGA25
 ENDIF
ENDIF
@ 3,0 CLEAR
DO HOTKEY WITH 'UNREGISTER','F'
DO HOTKEY WITH 'UNREGISTER','Q'
IF bPRINT
 DO HOTKEY WITH 'UNREGISTER','P'
ENDIF
CLEAR TYPEAHEAD
DO CASE
 CASE cFUNC = 'P' .AND. PRINTER()
  DO PLSWAIT
  cALERT_WIN = ''
  DO ALERT WITH 'NOTE','',-1,-1,'Printing ... press [Esc]', ;
                                '  if you want to abort.'
  nHANDLE = FOPEN(cOUTFILE)
  SET CONSOLE OFF
  SET PRINT ON
  _bESC = .f.
  ON ESCAPE _bESC = .t.
  SET ESCAPE ON
  ??? cSTART
  DO WHILE (.NOT. FEOF(nHANDLE)) .AND. (.NOT. _bESC)
   c = FGETS(nHANDLE)
   ??? c + _cEOL
  ENDDO
  SET ESCAPE OFF
  ON ESCAPE
  IF _bESC
   ? '*** ABORTED BY OPERATOR ***'
  ELSE
   ??? cEND
  ENDIF
  SET PRINT OFF
  SET CONSOLE ON
  = FCLOSE(nHANDLE)
  ERASE (cOUTFILE)
  RELEASE WINDOW alert
  DO ALERT WITH 'NOTE','OK',-1,-1,'Printing complete.'
  RETURN
 CASE cFUNC = 'F'
  DO WHILE .t.
   cFILE = ;
     RTRIM(DIALOG('C(35)','',-1,-1,'File this report under what file name?'))
   IF EXISTS(cFILE)
    _cCONTEXT = 'FileExist'
    c = ALERT('CAUTION','yNc',-1,-1,[File "] + cFILE ;
            + [" already exists.  Replace it?])
    _cCONTEXT = ''
    IF c = 'N'
     LOOP
    ENDIF
    IF c = 'C'
     ERASE (cOUTFILE)
     RETURN
    ENDIF
    ERASE (cFILE)
   ENDIF
   cNEWPATH = cFILE
   cOLDPATH = cOUTFILE
   IF .NOT. (':' $ cNEWPATH .OR. '\' $ cNEWPATH)
    cNEWPATH = FULLPATH(cNEWPATH)
   ENDIF
   IF .NOT. (':' $ cOLDPATH .OR. '\' $ cOLDPATH)
    cOLDPATH = FULLPATH(cOLDPATH)
   ENDIF
   IF '\' $ cNEWPATH
    n = RAT('\',cNEWPATH) - 1
    IF n = 0
     cNEWPATH = '\'
    ELSE
     cNEWPATH = LEFT(cNEWPATH,RAT('\',cNEWPATH) - 1)
    ENDIF
   ELSE
    cNEWPATH = LEFT(cNEWPATH,2)
   ENDIF
   IF '\' $ cOLDPATH
    n = RAT('\',cOLDPATH) - 1
    IF n = 0
     cOLDPATH = '\'
    ELSE
     cOLDPATH = LEFT(cOLDPATH,RAT('\',cOLDPATH) - 1)
    ENDIF
   ELSE
    cOLDPATH = LEFT(cOLDPATH,2)
   ENDIF
   IF cNEWPATH = cOLDPATH
    RENAME (cOUTFILE) TO (cFILE)
   ELSE
    COPY FILE (cPRINTTMP) TO (cFILE)
    ERASE (cOUTFILE)
   ENDIF
   EXIT
  ENDDO
  _cCONTEXT = 'Filed'
  DO ALERT WITH 'NOTE','OK',-1,-1,[Printout filed as "] + cFILE + ["]
  _cCONTEXT = ''
  RETURN
 OTHERWISE
  ERASE (cOUTFILE)
  RETURN
ENDCASE
RETURN

FUNCTION DISPEXIT
* Key interrupt attached to P,F & Q - exit the editor by stuffing keyboard
* buffer with [Ctrl-W]; return one of these 3 keys.
KEYBOARD CHR(23)
RETURN UPPER(CHR(LASTKEY()))

FUNCTION VMODE_TOG
* Key interrupt attached to ALT-V - if EGA or VGA, toggles between 25 lines
* and 43 or 50 line text modes, if available.
* Requires PUBLIC _cERRHAND contain program and arguments for error handler.
* Always returns new # of lines as number and stuffs keyboard with [Ctrl-W]
* if a mode toggle actually takes place.
PRIVATE c
* Make sure we have EGA or VGA
c = UPPER(SYS(2006))
IF .NOT. ('VGA' $ c .OR. 'EGA' $ c)
 RETURN 25
ENDIF
KEYBOARD CHR(23)
IF SROWS() = 25
 * Currently in 25 line mode, need to toggle to highest mode supported.
 * This may generate an error 216 (display mode not available) if, for
 * instance, QEMM's VIDRAM or similar utility is stealing graphics memory
 * for other purposes; so, we trap this error and ignore it if it occurs
 * by eating up our own [Ctrl-W].
 n = 0
 ON ERROR n = ERROR()
 IF 'EGA' $ c
  SET DISPLAY TO EGA43
 ELSE
  SET DISPLAY TO VGA50
 ENDIF
 ON ERROR DO &_cERRHAND
 DO CASE
  CASE n = 0
   RETURN IIF('EGA' $ c,43,50)
  CASE n = 216
   = INKEY()
   RETURN 25
  OTHERWISE
   DO ERRHAND WITH n,'Unexpected error','VMODE_TOG()','',0
   RETURN 25
 ENDCASE
ELSE
 * In extended text mode, return to standard 25 lines.
 IF 'EGA' $ c
  SET DISPLAY TO EGA25
 ELSE
  SET DISPLAY TO VGA25
 ENDIF
 RETURN 25
ENDIF

PROCEDURE THERM
*
*    THIS PROCEDURE WILL DISPLAY A THERMOMETER, INDICATING THE PERCENTAGE
*    COMPLETED TO THE NEAREST 1%.
*
*    PARAMETERS:
*      nNUM      -  CURRENT COUNT
*      nMAX      -  MAXIMUM COUNT
*      cMSG1     -  OPTIONAL, MESSAGE LINE 1
*      cMSG2     -  OPTIONAL, MESSAGE LINE 2 (if used cMSG1 must also be used)
*    NOTE:  IF BOTH nNUM AND nMAX ARE ZERO, THERMOMETER IS RESET TO START
*
*    VARIABLES:
*      nROW      -  ROW POSITION
*      nCOL      -  COLUMN POSITION
*      nTEMP     -  EXACT NO. OF CHAR POSITIONS TO SHOW
*      bHALFPOS  -  TRUE IF WE NEED 1/2 CELL AT LAST POSITION
*      nFULLPOS  -  NO. OF FULL CHAR POSITIONS TO SHOW
*      nMSGLINES -  NO. OF MESSAGE LINES -- 0, 1 OR 2

PARAMETERS nNUM,nMAX,cMSG1,cMSG2
PRIVATE nTEMP, nFULLPOS, bHALFPOS, nMSGLINES
nMSGLINES = PARAMETERS() - 2
IF TYPE("nNUM") = "L" .AND. WEXIST('therm')
  RELEASE WINDOW therm
  RETURN
ENDIF
IF nNUM + nMAX = 0
 IF nMSGLINES > 0
  @ 0,0 SAY PADR(cMSG1,50)
 ENDIF
 IF nMSGLINES > 1
  @ 1,0 SAY PADR(cMSG2,50)
 ENDIF
 @ nMSGLINES,0
 RETURN
ENDIF
nNUM = MIN(nMAX,nNUM)
nTEMP = 50 * (ROUND(nNUM,0) / nMAX)
IF nTEMP >= INT(nTEMP) + .5
  nFULLPOS = ROUND(nTEMP,0) - 1
  bHALFPOS = .f.
ELSE
  nFULLPOS = nTEMP - 1
  bHALFPOS = .t.
ENDIF
IF .NOT. WEXIST('therm')
  PRIVATE nCOL,nROW
  nCOL = ROUND((SCOLS()-1)/2,0)
  nROW = ROUND((SROWS()-1)/2,0)
  IF nMSGLINES = 2
   nROW = nROW - 1
  ENDIF
  DEFINE WINDOW therm FROM nROW - 1,nCOL-26 TO nROW + 3 + nMSGLINES,nCOL + 26 ;
                  FLOAT SHADOW TITLE " Percentage Completed " COLOR SCHEME 5
  ACTIVATE WINDOW therm
  @ 1 + nMSGLINES,1 SAY '....|....|....|....|....|....|....|....|....|....'
  @ 2 + nMSGLINES,4 SAY    '10   20   30   40   50   60   70   80   90'
ENDIF
IF .NOT. WOUTPUT() = 'THERM'
 ACTIVATE WINDOW therm
ENDIF
IF nMSGLINES > 0
 @ 0,0 SAY PADR(cMSG1,50)
ENDIF
IF nMSGLINES > 1
 @ 1,0 SAY PADR(cMSG2,50)
ENDIF
IF nFULLPOS >= 0
   @ nMSGLINES,0 SAY ;
           CHR(222) + REPLICATE(CHR(219),nFULLPOS) + IIF(bHALFPOS,CHR(221),'')
ENDIF
RETURN

FUNCTION CALLER
* This function returns the name of the program that called the routine that
* invokes this function.
PRIVATE n
* Find out how deep we are in the stack
n = 1
DO WHILE .t.
 IF PROGRAM(n) = 'CALLER'
  EXIT
 ENDIF
 n = n + 1
ENDDO
RETURN IIF(n < 2,'',PROGRAM(n - 2))

PROCEDURE LOOKUP_
bLOOKUP = .t.
KEYBOARD CHR(13)
RETURN
